/* Bluefish HTML Editor
 * document.c - this file contains the document code
 *
 * Copyright (C) 2000 Olivier Sessink
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "default_include.h"

#include <locale.h>

#include <stdio.h>				/* fopen */
#include <string.h>				/*strchr, memset */
#include <stdlib.h>				/* atoi */
#include <unistd.h>				/* unlink() */
#include <dirent.h>				/* DIR struct, open_dir() */
#include <time.h>				/* time() */

#include "bluefish.h"
#include "document.h"
#include "gtk_easy.h"			/* error dialog */
#include "interface.h"			/* statusbar message */
#include "undo.h"				/*undo_list_add */
#include "highlight.h"			/*line_hi_cb */
#include "rpopup.h"				/* rpopup_eventh */
#include "stringlist.h"			/* free_stringlist */
#include "snr2.h"				/* find_filenames in test_link_management() */
#include "menu.h"				/* add_to_recent_list */

/*******************************************************************/

typedef struct {
	GtkWidget *print_window;	/* the print dialog window */
	GtkWidget *print_cmd;		/* this will hold the command to be executed */
	GtkWidget *temp_dir;		/* this holds the temporary directory, e.g. /tmp */
} Tprintwin;

int pos_idx;

void setpos(gint position)
{
	gtk_editable_set_position(GTK_EDITABLE
							  (main_v->current_document->textbox),
							  position);
	gtk_timeout_remove(pos_idx);
}

/*******************************************************************/


/*******************************************************************/
/* set several document settings */

void document_set_fonts(gchar * textfontstring, gchar * labelfontstring,
						Tdocument * document)
{
	if (labelfontstring) {
		apply_font_style(document->textbox, textfontstring);
	}
	if (labelfontstring) {
		apply_font_style(document->tab_label, labelfontstring);
	}
}

void document_set_wrap(gint word_wrap, gint line_wrap,
					   Tdocument * document)
{
	gtk_text_set_line_wrap(GTK_TEXT(document->textbox), line_wrap);
	gtk_text_set_word_wrap(GTK_TEXT(document->textbox), word_wrap);
}

/*************************************************************************/

static gboolean doc_highlight_line_once(Tdocument * whichdoc)
{
	DEBUG_MSG("highlight_line_once, need_highlighting=%d\n",
			  whichdoc->need_highlighting);
	whichdoc->need_highlighting = 0;
	line_hi_cb(NULL, whichdoc);
	return FALSE;
}

static gboolean doc_highlight_once(Tdocument * whichdoc)
{
	DEBUG_MSG("doc_highlight_once, need_highlighting=%d\n",
			  whichdoc->need_highlighting);
	whichdoc->need_highlighting = 0;
	refresh_hi_cb(NULL, whichdoc);
	return FALSE;
}

void doc_need_highlighting(Tdocument * whichdoc)
{
	DEBUG_MSG
		("doc_need_highlighting, need_highlighting=%d, cont_highlight_full=%d\n",
		 whichdoc->need_highlighting, main_v->props.cont_highlight_full);
	if (whichdoc->highlightstate && main_v->props.cont_highlight_update) {
		if (main_v->props.cont_highlight_full) {
			gtk_timeout_add(100, (GtkFunction) doc_highlight_once,
							whichdoc);
		} else {
			gtk_timeout_add(100, (GtkFunction) doc_highlight_line_once,
							whichdoc);
		}
	}
}

/************************************************************************/

gint find_filename_in_documentlist(gchar *filename) {
	GList *tmplist;
	gint count=0;

	if (!filename) {
		return -1;
	}
	
	tmplist = g_list_first(main_v->documentlist);
	while (tmplist) {
		DEBUG_MSG("test_doc_already_open, tmplist=%p\n", tmplist);
		if (((Tdocument *)tmplist->data)->filename &&(strcmp(filename, ((Tdocument *)tmplist->data)->filename) ==0)) {
			return count;
		}
		count++;
		tmplist = g_list_next(tmplist);
	}
	return -1;
}

gint test_docs_modified(GList *doclist) {

	GList *tmplist;
	Tdocument *tmpdoc;

	if (doclist) {
		tmplist = g_list_first(doclist);
	} else {
		tmplist = g_list_first(main_v->documentlist);
	}
	
	while (tmplist) {
		tmpdoc = (Tdocument *) tmplist->data;
#ifdef DEBUG
		g_assert(tmpdoc);
#endif
		if (tmpdoc->modified) {
			DEBUG_MSG("test_docs_modified, found a modified document\n");
			return 1;
		}
		tmplist = g_list_next(tmplist);
	}
	DEBUG_MSG("test_docs_modified, no modified documents found\n");
	return 0;
}

gint test_only_empty_doc_left(Tdocument *doc) {
	Tdocument *tmpdoc;

	if (!doc) {
		tmpdoc = main_v->current_document;
	} else {
		tmpdoc = doc;
	}

	if (!tmpdoc->modified && !tmpdoc->filename
				&& (gtk_text_get_length(GTK_TEXT(tmpdoc->textbox)) == 0)
				&& (g_list_length(main_v->documentlist) == 1)) {
			DEBUG_MSG("test_only_empty_doc_left, returning 1\n");
			return 1;
	}
	DEBUG_MSG("test_only_empty_doc_left, returning 0\n");
	return 0;
}


/*
 * PUBLIC: doc_insert_text_cb, some parts copied from doc.c in the gnotepad code
 *
 * callback for text widget for text insertion from keyboard typing.  note that
 * this routine and the next one (doc_delete_text_cb()), gets invoked for EVERY
 * key typed by the user.
 * also cotains some highlighting related stuff and an autoindenter.
 */
void doc_insert_text_cb(GtkEditable * editable, const char *text, int len, int
						*pos, gpointer d)
{
	char *buf, space;
	gint newlines = 0, position, count, touch = 0, max;


	if (main_v->props.cont_highlight_update) {
		if ((text[0] == '"') || (text[0] == '<') || (text[0] == '>')) {
			doc_need_highlighting(main_v->current_document);
		}
	}
	/* does this introduce a memory leak? where is this buf free-ed? 
	   I hope undo_list_add() uses it to store the change, but this is not verified */
	buf = g_strndup(text, len);



	/* we need the current position several times ;) */
	position =
		gtk_editable_get_position(GTK_EDITABLE
								  (main_v->current_document->textbox));
	count = position;
	max = position;

	if ((main_v->props.autoindent == 1) && (text[0] == '\n')) {
		/* don't count that newline you just entered  */
		position--;

		/* go back till you find newlines. count the chars in max, we'll need it later */
		while ((newlines < 1) && (position > 0)) {
			if (GTK_TEXT_INDEX
				(GTK_TEXT(main_v->current_document->textbox),
				 position) == '\n') {
				newlines++;
			}
			position--;
			max++;
		}
		/* one for the last go and one cause we need the char after the newline */
		position = position + 2;


		space =
			GTK_TEXT_INDEX(GTK_TEXT(main_v->current_document->textbox),
						   position);

		/* insert the stuff - don't do more than the max we counted before - this is because if space=='\n', this
		 * statement is true. don't ask me why. 
		 */
		while (((space == '\t') || (space == ' ')) && (count < max - 1)) {
			gtk_text_insert(GTK_TEXT(main_v->current_document->textbox),
							NULL, NULL, NULL, &space, 1);
			position++;
			space =
				GTK_TEXT_INDEX(GTK_TEXT(main_v->current_document->textbox),
							   position);
			count++;
			touch = 1;
		}

		if (touch == 1) {
			count++;

			/* this is real ugly. it seems to be a bug in gtk, you cannot set the position when
			 * you just entered a newline. so i build a workaround, with gtk_timeout_add, as you can see.
			 */
			pos_idx =	gtk_timeout_add(10, (GtkFunction) setpos,	(gpointer) count);
		}
	}


	DEBUG_MSG("doc_insert_text_cb, buf=%s\n", buf);
	undo_list_add(d, buf, *pos, *pos + len, UndoInsert);
	main_v->current_document->modified = 1;
	DEBUG_MSG
		("doc_insert_text_cb, finished, modified = 1, need_highlighting=%d\n",
		 main_v->current_document->need_highlighting);
}


								/* doc_insert_text_cb */


/*
 * PUBLIC: doc_delete_text_cb, copied from doc.c in the gnotepad code
 *
 * callback for text widget for text deletion from keyboard typing
 */
void doc_delete_text_cb(GtkEditable * editable, int start, int end,
						gpointer d)
{
	char *text;
	DEBUG_MSG
		("doc_delete_text_cb, started, editable=%p, start=%d, end=%d, document=%p\n",
		 editable, start, end, d);
	text = gtk_editable_get_chars(editable, start, end);
	DEBUG_MSG("doc_delete_text_cb, text=%p, document->textbox=%p\n", text,
			  ((Tdocument *) d)->textbox);
	undo_list_add(d, text, start, end, UndoDelete);
	main_v->current_document->modified = 1;
	DEBUG_MSG("doc_delete_text_cb, finished, modified=1\n");
}								/* doc_delete_text_cb */

Ttext_positions get_positions(GtkWidget * textbox)
{

	Ttext_positions positions;
	gint tmp;

	positions.currentp = gtk_editable_get_position(GTK_EDITABLE(textbox));
	positions.adj = GTK_TEXT(textbox)->vadj->value;
	positions.selection = GTK_EDITABLE(textbox)->has_selection;
	if (positions.selection) {
		positions.startp = GTK_EDITABLE(textbox)->selection_start_pos;
		positions.endp = GTK_EDITABLE(textbox)->selection_end_pos;
		if (positions.endp < positions.startp) {
			tmp = positions.endp;
			positions.endp = positions.startp;
			positions.startp = tmp;
		}
	} else {
		positions.endp = positions.currentp;
		positions.startp = positions.currentp;
	}
	return positions;
}

void restore_positions(Ttext_positions positions, GtkWidget * textbox)
{

	gtk_editable_set_position(GTK_EDITABLE(textbox), positions.currentp);
	gtk_adjustment_set_value(GTK_TEXT(textbox)->vadj, positions.adj);
	if (positions.selection) {
		DEBUG_MSG("restore_positions, startp=%d, endp=%d\n",
				  positions.startp, positions.endp);
		gtk_editable_select_region(GTK_EDITABLE(textbox), positions.startp,
								   positions.endp);
	}
}


void doc_replace_text(const gchar * newstring, gint pos, gint end,
					  Tdocument * doc)
{
#ifdef DEBUG
	gchar *tmpstring;
	gint oldlen, newlen;
#endif
	gtk_text_freeze(GTK_TEXT(doc->textbox));
	DEBUG_MSG
		("doc_replace_text, started, pos=%d, end=%d, newstring=%s, strlen(newstring)=%d\n",
		 pos, end, newstring, strlen(newstring));
#ifdef DEBUG
	tmpstring =
		gtk_editable_get_chars(GTK_EDITABLE(GTK_TEXT(doc->textbox)), pos,
							   end);
	if (!tmpstring) {
		DEBUG_MSG
			("*** doc_replace_text, PROBLEM: tmpstring=NULL ***\nnewstring=%s, pos=%d, end=%d, doc=%p\n",
			 newstring, pos, end, doc);
		return;
	}
	DEBUG_MSG("doc_replace_text, about to replace %s\n", tmpstring);
	oldlen = strlen(tmpstring);
	newlen = strlen(newstring);
	DEBUG_MSG("doc_replace_text, oldlen=%d, newlen=%d, change=%d\n",
			  oldlen, newlen, newlen - oldlen);
#endif
	gtk_editable_delete_text(GTK_EDITABLE(GTK_TEXT(doc->textbox)), pos,
							 end);
	DEBUG_MSG("doc_replace_text, text deleted\n");
	gtk_editable_insert_text(GTK_EDITABLE(GTK_TEXT(doc->textbox)),
							 newstring, strlen(newstring), &pos);
	gtk_text_thaw(GTK_TEXT(doc->textbox));
	doc->modified = 1;
	doc_need_highlighting(doc);
}

void replace_text(const gchar * newstring, gint pos, gint end)
{
	DEBUG_MSG
		("replace_text, calling doc_replace_text on current_document\n");
	doc_replace_text(newstring, pos, end, main_v->current_document);
}

void doc_insert_text(Tdocument * document, const gchar * before,
					 const gchar * after)
{
	guint pos;
	Ttext_positions positions;

	if ((!before && !after) || !document)
		return;

	/* find the current cursor position and selection position */
	positions = get_positions(document->textbox);

/* freeze the textbox, this is better when inserting large blocks */
	gtk_text_freeze(GTK_TEXT(document->textbox));

	DEBUG_MSG("insert_text, A currentp = %d, startp=%d, endp=%d\n",
			  positions.currentp, positions.startp, positions.endp);
	if (after != NULL && strlen(after) > 0) {
		gtk_text_set_point(GTK_TEXT(document->textbox), positions.endp);
		gtk_text_insert(GTK_TEXT(document->textbox), NULL, NULL, NULL,
						after, -1);
		/* strlen() is unable to find the lenght of multibyte characters (japanese)
		   so we use the new position in the text box to find out */
		pos = gtk_text_get_point(GTK_TEXT(document->textbox));
		undo_list_add(document, g_strdup(after), positions.endp, pos,
					  UndoInsert);
	} else {
		pos = positions.endp;
	}

	if (before != NULL && strlen(before) > 0) {
		gtk_text_set_point(GTK_TEXT(document->textbox), positions.startp);
		gtk_text_insert(GTK_TEXT(document->textbox), NULL, NULL, NULL,
						before, -1);
		pos = gtk_text_get_point(GTK_TEXT(document->textbox));
		undo_list_add(document, g_strdup(before), positions.startp, pos,
					  UndoInsert);
	} else {
		pos = positions.startp;
	}

/* thaw textbox, we're finished */
	DEBUG_MSG("insert_text, ->thaw\n");
	gtk_text_thaw(GTK_TEXT(document->textbox));
	positions.endp += pos - positions.startp;
	positions.currentp += pos - positions.startp;
	positions.startp = pos;
	restore_positions(positions, document->textbox);

	/* grab focus so we can continue working */
	DEBUG_MSG("insert_text, about to grab focus\n");
	gtk_widget_grab_focus(GTK_WIDGET(document->textbox));
	DEBUG_MSG("insert_text, set modified=1\n");
	document->modified = 1;

	doc_need_highlighting(document);
}

void insert_text(const gchar * before, const gchar * after)
{
	doc_insert_text(main_v->current_document, before, after);
}


static void scroll_textbox(gint direction_up)
{
	gfloat newvalue;

	newvalue =
		GTK_ADJUSTMENT(GTK_TEXT(main_v->current_document->textbox)->vadj)->
		value;
	if (direction_up) {
		newvalue +=
			(GTK_ADJUSTMENT
			 (GTK_TEXT(main_v->current_document->textbox)->vadj)->
			 page_increment / 2);
	} else {
		newvalue -=
			(GTK_ADJUSTMENT
			 (GTK_TEXT(main_v->current_document->textbox)->vadj)->
			 page_increment / 2);
	}
	gtk_adjustment_set_value(GTK_ADJUSTMENT
							 (GTK_TEXT(main_v->current_document->textbox)->
							  vadj), newvalue);
}

static gint textbox_event_lcb(GtkWidget * widget, GdkEvent * event)
{

	if (event->type == GDK_BUTTON_PRESS) {
		GdkEventButton *bevent = (GdkEventButton *) event;
		DEBUG_MSG("rpopup_eventh, bevent->button=%d\n", bevent->button);
		if (bevent->button == 3) {
			rpopup_eventh(bevent);
			return TRUE;
		} else if (bevent->button == 4) {
			scroll_textbox(0);
			return TRUE;
		} else if (bevent->button == 5) {
			scroll_textbox(1);
			return TRUE;
		}
	}
	/* Tell calling code that we have not handled this event; pass it on. */
	return FALSE;
}


/* static void scrollbar_update(GtkText * txt, gint tl, gint ln)
{
	gfloat value;

	if (tl < 3 || ln > tl)
		return;

	value = (ln * GTK_ADJUSTMENT(txt->vadj)->upper) / tl - GTK_ADJUSTMENT(txt->vadj)->page_increment;

	gtk_adjustment_set_value(GTK_ADJUSTMENT(txt->vadj), value);
} */

gint position_to_linenum(gint position)
{
	gint i, lines;
	gchar *c;					/* = g_malloc0(3); */

	lines = 0;
	i = gtk_text_get_length(GTK_TEXT(main_v->current_document->textbox));
	if (position > i) {
		position = i;
	}
	c =
		gtk_editable_get_chars(GTK_EDITABLE
							   (main_v->current_document->textbox), 0,
							   position);
	if (!c)
		return 0;
	DEBUG_MSG("position_to_linenum, strlen(c)=%d, position=%d\n",
			  strlen(c), position);
	/* for multibyte-text support, don't use 'position' */
	for (i = 0; c[i]; i++) {
		if (c[i] == '\n')
			lines++;
	}
	g_free(c);
	return lines;
}

void go_to_line(gint linenum, gint select_line)
{
/* Thanks to Andy Kahn for the amazing non-leaking go to line function
   It dances, it does tricks, AND you have memory left after... :) */

	gchar *buf, *haystack, *needle;
	gint numlines;
	gint a, b, len;

	DEBUG_MSG("doc_go_to_line, started, linenum=%d, select_line=%d\n",
			  linenum, select_line);
	numlines = 1;

	len = gtk_text_get_length(GTK_TEXT(main_v->current_document->textbox));

/*  If index is past the end, or 0 (Usually because line number is -) */

	if ((linenum > (len - 1)) || (linenum == 0))
		return;

	a = 0;
	b = len;
	if (GTK_TEXT(main_v->current_document->textbox)->use_wchar) {
		int i;
		GtkText *text = GTK_TEXT(main_v->current_document->textbox);
		for (i = 0; i < len; i++) {
			if ((int) GTK_TEXT_INDEX(text, i) == '\n') {
				if (linenum == numlines) {
					b = i;
					break;
				}
				numlines++;
				if (linenum == numlines)
					a = i + 1;
			}
		}
	} else {
		buf =
			gtk_editable_get_chars(GTK_EDITABLE
								   (main_v->current_document->textbox), 0,
								   len);
		haystack = buf;
		do {
			needle = strchr(haystack, '\n');
			if (needle) {
				haystack = needle + 1;
				if (linenum == numlines) {
					b = needle - buf;
					break;
				}
				numlines++;
				if (linenum == numlines)
					a = needle - buf + 1;
			}
		}
		while (needle != NULL);

		g_free(buf);
	}
	DEBUG_MSG("doc_go_to_line, started, a=%d, b=%d\n", a, b);

	if (select_line) {
		gtk_editable_select_region(GTK_EDITABLE
								   (main_v->current_document->textbox), a,
								   b);
	}

/* this oneliner does the trick as well :) */

	gtk_editable_set_position(GTK_EDITABLE
							  (GTK_TEXT
							   (main_v->current_document->textbox)), b);


/*	DEBUG_MSG("doc_go_to_line, going to edit/delete a char to position the widget\n");	
	gtk_signal_disconnect(GTK_OBJECT(main_v->current_document->textbox), main_v->current_document->ins_txt_id);
	gtk_signal_disconnect(GTK_OBJECT(main_v->current_document->textbox), main_v->current_document->del_txt_id); 

	gtk_editable_insert_text(GTK_EDITABLE((GtkText *)
										  main_v->current_document->textbox), " ", 1, &b);
	gtk_editable_delete_text(GTK_EDITABLE((GtkText *)
										  main_v->current_document->textbox), b - 1, b);

	main_v->current_document->ins_txt_id = gtk_signal_connect(GTK_OBJECT(main_v->current_document->textbox), "insert_text", GTK_SIGNAL_FUNC(doc_insert_text_cb), main_v->current_document);
	main_v->current_document->del_txt_id = gtk_signal_connect(GTK_OBJECT(main_v->current_document->textbox), "delete_text", GTK_SIGNAL_FUNC(doc_delete_text_cb), main_v->current_document);
*/
}


/*****************************************************/

void copy_cb(GtkWidget * w, gpointer data)
{
	gtk_editable_copy_clipboard(GTK_EDITABLE
								(main_v->current_document->textbox));
}

/*****************************************************/

void paste_cb(GtkWidget * w, gpointer data)
{
	gtk_editable_paste_clipboard(GTK_EDITABLE
								 (main_v->current_document->textbox));
	main_v->current_document->modified = 1;
	if (main_v->current_document->highlightstate) {
		doc_need_highlighting(main_v->current_document);
	}
}


/*****************************************************/

void cut_cb(GtkWidget * w, gpointer data)
{
	gtk_editable_cut_clipboard(GTK_EDITABLE
							   (main_v->current_document->textbox));
	main_v->current_document->modified = 1;
	if (main_v->current_document->highlightstate) {
		doc_need_highlighting(main_v->current_document);
	}
}

/*****************************************************/

void sel_all_cb(GtkWidget * w, gpointer data)
{
	gtk_editable_select_region(GTK_EDITABLE
							   (main_v->current_document->textbox), 0,
							   gtk_text_get_length(GTK_TEXT
												   (main_v->
													current_document->
													textbox)));
}

/*****************************************************/

void all_documents_update_links(Tdocument * curdoc, gchar * oldfilename,
								gchar * newfilename)
{
	gchar *eff_newfilename, *eff_oldfilename;
	GList *tmplist;
	gint count =
		bf_statusbar_message(_("Updating Links in all Documents"));

	eff_newfilename = most_efficient_filename(g_strdup(newfilename));
	eff_oldfilename = most_efficient_filename(g_strdup(oldfilename));
	tmplist = g_list_first(main_v->documentlist);
	while (tmplist) {
		if ((Tdocument *) tmplist->data != curdoc) {
			update_filenames_in_file((Tdocument *) tmplist->data,
									 eff_oldfilename, eff_newfilename, 0);
		}
		tmplist = g_list_next(tmplist);
	}
	g_free(eff_newfilename);
	g_free(eff_oldfilename);
	statusbar_remove(GINT_TO_POINTER(count));
}

/**********************************************/



static void go_to_line_lcb(GtkWidget * widget, gpointer data)
{
	gchar *linestr;
	gint linenum;
	GtkWidget *entry;

	if (data) {
		entry = (GtkWidget *) data;
		linestr = gtk_editable_get_chars(GTK_EDITABLE(entry), 0, -1);
		linenum = atoi(linestr);
		DEBUG_MSG("go_to_line_cb, going to line %d (string:%s)\n", linenum,
				  linestr);
		go_to_line(linenum, 1);
	}
	gtk_widget_destroy(GTK_WIDGET(gtk_widget_get_toplevel(widget)));
}


void go_to_line_win_cb(GtkWidget * widget, gpointer data)
{
	GtkWidget *win, *table, *entry;
	GtkWidget *but1, *but2;
	win =
		window_with_title(_("Goto line"), GTK_WIN_POS_MOUSE,
						  GTK_WINDOW_DIALOG, 5);

	table = gtk_table_new(3, 2, 0);
	gtk_container_add(GTK_CONTAINER(win), table);

	entry = gtk_entry_new();
	gtk_table_attach_defaults(GTK_TABLE(table),
							  gtk_label_new(_("Line number: ")), 0, 2, 0,
							  1);
	gtk_table_attach_defaults(GTK_TABLE(table), entry, 0, 2, 1, 2);
	but1 = bf_stock_ok_button(go_to_line_lcb, entry);
	but2 = bf_stock_cancel_button(go_to_line_lcb, NULL);
	gtk_table_attach_defaults(GTK_TABLE(table), but1, 0, 1, 2, 3);
	gtk_table_attach_defaults(GTK_TABLE(table), but2, 1, 2, 2, 3);
	gtk_widget_show_all(win);
}

typedef struct {
	char id;
	char *entity;
} Tchar_entity;

static Tchar_entity ascii_charset[] = {
	{34, "&quot;"}, {38, "&amp;"}, {60, "&lt;"}, {62, "&gt;"}, {0, NULL}
};

static Tchar_entity iso8859_1_charset[] = {
	{160, "&nbsp;"}, {161, "&iexcl;"}, {162, "&cent;"}, {163, "&pound;"},
	{164, "&curren;"}, {165, "&yen;"}, {166, "&brvbar;"}, {167, "&sect;"},
	{168, "&uml;"}, {169, "&copy;"}, {170, "&ordf;"}, {171, "&laquo;"},
	{172, "&not;"}, {173, "&shy;"}, {174, "&reg;"}, {175, "&macr;"},
	{176, "&deg;"}, {177, "&plusmn;"}, {178, "&sup2;"}, {179, "&sup3;"},
	{180, "&acute;"}, {181, "&micro;"}, {182, "&para;"}, {183, "&middot;"},
	{184, "&cedil;"}, {185, "&sup1;"}, {186, "&ordm;"}, {187, "&raquo;"},
	
		{188, "&frac14;"}, {189, "&frac12;"}, {190, "&frac34;"}, {191,
																  "&iquest;"},
	{192, "&Agrave;"}, {193, "&Aacute;"}, {194, "&Acirc;"}, {195,
															 "&Atilde;"},
	{196, "&Auml;"}, {197, "&Aring;"}, {198, "&AElig;"}, {199, "&Ccedil;"},
	{200, "&Egrave;"}, {201, "&Eacute;"}, {202, "&Ecirc;"}, {203,
															 "&Euml;"},
	{204, "&Igrave;"}, {205, "&Iacute;"}, {206, "&Icirc;"}, {207,
															 "&Iuml;"},
	{208, "&ETH;"}, {209, "&Ntilde;"}, {210, "&Ograve;"}, {211,
														   "&Oacute;"},
	{212, "&Ocirc;"}, {213, "&Otilde;"}, {214, "&Ouml;"}, {215,
														   "&times;"},
	{216, "&Oslash;"}, {217, "&Ugrave;"}, {218, "&Uacute;"}, {219,
															  "&Ucirc;"},
	{220, "&Uuml;"}, {221, "&Yacute;"}, {222, "&THORN;"}, {223,
														   "&szlig;"},
	{224, "&agrave;"}, {225, "&aacute;"}, {226, "&acirc;"}, {227,
															 "&atilde;"},
	{228, "&auml;"}, {229, "&aring;"}, {230, "&aelig;"}, {231,
														  "&ccedil;"},
	{232, "&egrave;"}, {233, "&eacute;"}, {234, "&ecirc;"}, {235,
															 "&euml;"},
	{236, "&igrave;"}, {237, "&iacute;"}, {238, "&icirc;"}, {239,
															 "&iuml;"},
	{240, "&eth;"}, {241, "&ntilde;"}, {242, "&ograve;"}, {243,
														   "&oacute;"},
	{244, "&ocirc;"}, {245, "&otilde;"}, {246, "&ouml;"}, {247,
														   "&divide;"},
	{248, "&oslash;"}, {249, "&ugrave;"}, {250, "&uacute;"}, {251,
															  "&ucirc;"},
	{252, "&uuml;"}, {253, "&yacute;"}, {254, "&thorn;"}, {255,
														   "&yuml;"}
	, {0, NULL}
};

/* charset 0 = ascii, 1 = iso8859-1 */

static void doc_special_char_replace(Tdocument * document, gint startpos,
									 gint endpos, gint charset)
{
	gint i, j, contj;
	char curchar;
	Tchar_entity *set;

	if (endpos == -1) {
		endpos = gtk_text_get_length(GTK_TEXT(document->textbox));
	}

	if (endpos < startpos) {
		i = startpos;
		startpos = endpos;
		endpos = i;
	}
	if (charset == 1) {
		set = &iso8859_1_charset[0];
	} else {
		set = &ascii_charset[0];
	}
	DEBUG_MSG
		("doc_special_char_replace, startpos=%d, endpos=%d, length=%d\n",
		 startpos, endpos,
		 gtk_text_get_length(GTK_TEXT(document->textbox)));
	gtk_text_freeze(GTK_TEXT(document->textbox));
	for (i = startpos; i <= endpos; i++) {
		curchar = (GTK_TEXT_INDEX(GTK_TEXT(document->textbox), i));
		DEBUG_MSG("doc_special_char_replace, i=%d, curchar=%d\n", i,
				  curchar);
		if (!curchar) {
			DEBUG_MSG("doc_special_char_replace, char %d doesn exist\n",
					  i);
			gtk_text_thaw(GTK_TEXT(document->textbox));
			return;
		}
		j = 0;
		contj = (set[j].entity != NULL);
		while (contj) {
			if (set[j].id == curchar) {
				gtk_text_set_point(GTK_TEXT(document->textbox), i);
				gtk_text_forward_delete(GTK_TEXT(document->textbox), 1);
				gtk_text_insert(GTK_TEXT(document->textbox), NULL, NULL,
								NULL, set[j].entity,
								strlen(set[j].entity));
				endpos += strlen(set[j].entity) - 1;
				contj = 0;
			} else {
				j++;
				contj = (set[j].entity != NULL);
			}
		}
	}
	gtk_text_thaw(GTK_TEXT(document->textbox));
	DEBUG_MSG("doc_special_char_replace, finished\n");
}


static void special_char_replace(gint charset)
{
	gint start, end, count;
	GtkWidget *textbox = main_v->current_document->textbox;

	count = bf_statusbar_message(_("Charset conversion running..."));

	if (GTK_EDITABLE(textbox)->has_selection) {
		start = GTK_EDITABLE(textbox)->selection_start_pos;
		end = GTK_EDITABLE(textbox)->selection_end_pos;
	} else {
		start = 0;
		end = -1;
	}
	doc_special_char_replace(main_v->current_document, start, end,
							 charset);
	if (main_v->current_document->highlightstate) {
		doc_highlight_once(main_v->current_document);
	}
	statusbar_remove(GINT_TO_POINTER(count));
}

void iso8859_1_replace_cb(GtkWidget * widget, gpointer data)
{
	special_char_replace(1);
}

void ascii_replace_cb(GtkWidget * widget, gpointer data)
{
	special_char_replace(0);
}


/*****************************************************/
/**** doc new, close, open, save etc. functions ******/
/*****************************************************/

void doc_file_to_textbox(Tdocument * doc, gchar * filename)
{
	FILE *fd;
	gchar *errmessage, line[512], *linebuff = NULL;
	gint buffsize = 0, newsize;

	statusbar_message(_("Opening file"), 1000);
	/* This opens the contents of a file to a textbox */
	change_dir(filename);
	fd = fopen(filename, "r");
	if (fd == NULL) {
		DEBUG_MSG("file_to_textbox, cannot open file %s\n", filename);
		errmessage =
			g_strconcat(_("Could not open file:\n"), filename, NULL);
		error_dialog(_("Error"), errmessage);	/* 7 */
		g_free(errmessage);
		return;
	} else {
		gtk_text_freeze(GTK_TEXT(doc->textbox));

		newsize = sizeof(line);
		linebuff = g_malloc(newsize);
		if (linebuff) {
			linebuff[0] = '\0';
			buffsize = newsize;
		}

		while (fgets(line, sizeof(line), fd) != NULL) {
			if (!linebuff) {
				gtk_text_insert(GTK_TEXT(doc->textbox),
					NULL, NULL, NULL, line, -1);
			} else {
				newsize = strlen(linebuff) + strlen(line) + 1;
				if (buffsize < newsize) {
					gchar *ptr = g_realloc( linebuff, newsize );
					if (!ptr) {
						DEBUG_MSG("realloc linebuff failed\n");
						gtk_text_insert(GTK_TEXT(doc->textbox),
							NULL, NULL, NULL, linebuff, -1);
						linebuff[0] = '\0';
						gtk_text_insert(GTK_TEXT(doc->textbox),
							NULL, NULL, NULL, line, -1);
						continue;
					}
					linebuff = ptr;	
					buffsize = newsize;
					DEBUG_MSG("realloc linebuff : %d bytes\n", newsize);
				}
				strcat( linebuff, line );
				if( strlen(linebuff)>0 &&
				    linebuff[strlen(linebuff)-1]=='\n' ) {
					gtk_text_insert(GTK_TEXT(doc->textbox),
						NULL, NULL, NULL, linebuff, -1);
					linebuff[0] = '\0';
				}
			}
		}
		if( linebuff ) {
			gtk_text_insert(GTK_TEXT(doc->textbox),
		 		NULL, NULL, NULL, linebuff, -1);
			g_free( linebuff );
			linebuff = NULL;
 		}
		gtk_text_thaw(GTK_TEXT(doc->textbox));
		fclose(fd);
		doc->modified = FALSE;
		if (doc->highlightstate) {
			doc_highlight_once(doc);
			DEBUG_MSG("file_to_textbox, %p needs highlighting\n", doc);
		}
	}
}

static gint file_check_backup(gchar * filename)
{
	gchar *backupfilename;
	gint res;

	if (main_v->props.backup_file && filename
		&& file_exists_and_readable(filename)) {
		backupfilename =
			g_strconcat(filename, main_v->props.backup_filestring, NULL);
		res = rename(filename, backupfilename);
		g_free(backupfilename);
		if (res == 0) {
			return 1;
		} else {
			return 0;
		}
	} else {
		return 1;
	}
}

void doc_textbox_to_file(Tdocument * doc, gchar * filename)
{
	FILE *fd;
	gchar *tmpchar, *errmessage = NULL;

	DEBUG_MSG("textbox_to_file, started file %s\n", filename);
	statusbar_message(_("Saving file"), 1000);
	/* This writes the contents of a textbox to a file */
	if (!file_check_backup(filename)) {
		DEBUG_MSG("textbox_to_file, could not make backup!!!\n");
		return;
	}

	change_dir(filename);
	fd = fopen(filename, "w");
	if (fd == NULL) {
		DEBUG_MSG("textbox_to_file, cannot open file %s\n", filename);
		errmessage =
			g_strconcat(_("Could not write file:\n"), filename, NULL);
		error_dialog(_("Error"), errmessage);
		g_free(errmessage);
		return;
	} else {
		DEBUG_MSG("textbox_to_file, lenght=%d\n",
				  gtk_text_get_length(GTK_TEXT(doc->textbox)));
		DEBUG_MSG("textbox_to_file, fd=%p\n", fd);
		DEBUG_MSG("textbox_to_file, filename=%s\n", filename);

		tmpchar =
			gtk_editable_get_chars(GTK_EDITABLE(doc->textbox), 0, -1);
		fputs(tmpchar, fd);
		g_free(tmpchar);
		fclose(fd);
		doc->modified = 0;
		DEBUG_MSG("textbox_to file, text saved to: %s\n", filename);
	}
}

/*************
file_save_cb is called in file_close_all_cb, but it shouldn't work on
the current document, so we need doc_save
also for file_save_as in file_save_all_cb
************/

Tdocument *doc_new(void)
{
	GtkWidget *tmptable, *tmpscrollbar;
	Tdocument *document;

	/* Here we create a new document, and add it to the GList */
	DEBUG_MSG("doc_new, started\n");
	if (main_v->current_document != NULL) {
		if(test_only_empty_doc_left(main_v->current_document)) {
			DEBUG_MSG("doc_new, no new needed, the current one is empty\n");
			return main_v->current_document;
		}
	}

	document = g_malloc0(sizeof(Tdocument));
	DEBUG_MSG("doc_new, malloced at %p\n", document);
	tmptable = gtk_table_new(2, 1, FALSE);
	gtk_table_set_row_spacings(GTK_TABLE(tmptable), 0);
	gtk_table_set_col_spacings(GTK_TABLE(tmptable), 0);
	document->textbox = gtk_text_new(NULL, NULL);
	DEBUG_MSG("doc_new, document->textbox=%p\n", document->textbox);
	gtk_table_attach_defaults(GTK_TABLE(tmptable), document->textbox, 0, 1,
							  0, 1);

	tmpscrollbar = gtk_vscrollbar_new(GTK_TEXT(document->textbox)->vadj);
	GTK_WIDGET_UNSET_FLAGS(tmpscrollbar, GTK_CAN_FOCUS);
	gtk_table_attach(GTK_TABLE(tmptable), tmpscrollbar, 1, 2, 0, 1,
					 GTK_FILL, GTK_EXPAND | GTK_FILL | GTK_SHRINK, 0, 0);
	gtk_widget_show(tmpscrollbar);

	document->tab_label = gtk_label_new(_("Untitled"));
	GTK_WIDGET_UNSET_FLAGS(document->tab_label, GTK_CAN_FOCUS);

	apply_font_style(GTK_WIDGET(document->tab_label),
					 main_v->props.cfg_tab_font);

	DEBUG_MSG("doc_new, before notebook_append_page\n");
	gtk_notebook_append_page(GTK_NOTEBOOK(main_v->notebook), tmptable,
							 document->tab_label);
	gtk_widget_show(tmptable);
	DEBUG_MSG("doc_new, before realize on textbox\n");
	gtk_widget_realize(document->textbox);
	DEBUG_MSG("doc_new, after realize on textbox\n");
	gtk_text_set_editable(GTK_TEXT(document->textbox), TRUE);
	document_set_wrap(main_v->props.word_wrap, main_v->props.line_wrap,
					  document);

	gtk_signal_connect_object(GTK_OBJECT(document->textbox), "event",
							  GTK_SIGNAL_FUNC(textbox_event_lcb), NULL);

	apply_font_style(GTK_WIDGET(document->textbox),
					 main_v->props.cfg_editor_font);
	GTK_TEXT(document->textbox)->default_tab_width =
		main_v->props.cfg_editor_tabwidth;
	GTK_TEXT(document->textbox)->tab_stops =
		g_list_remove(GTK_TEXT(document->textbox)->tab_stops,
					  GTK_TEXT(document->textbox)->tab_stops->data);
	GTK_TEXT(document->textbox)->tab_stops =
		g_list_remove(GTK_TEXT(document->textbox)->tab_stops,
					  GTK_TEXT(document->textbox)->tab_stops->data);
	GTK_TEXT(document->textbox)->tab_stops = NULL;
	GTK_TEXT(document->textbox)->tab_stops =
		g_list_prepend(GTK_TEXT(document->textbox)->tab_stops,
					   (void *) main_v->props.cfg_editor_tabwidth);
	GTK_TEXT(document->textbox)->tab_stops =
		g_list_prepend(GTK_TEXT(document->textbox)->tab_stops,
					   (void *) main_v->props.cfg_editor_tabwidth);

	gtk_widget_show(document->textbox);

	document->modified = FALSE;
	document->filename = NULL;
	document->need_highlighting = 0;
	document->highlightstate = main_v->props.defaulthighlight;

	main_v->documentlist = g_list_append(main_v->documentlist, document);
	gtk_widget_grab_focus(document->textbox);

	/* this part is copied from gnotepad code 
	   Copyright (C) 1998 Andy C. Kahn <kahn@zk3.dec.com> */

	document->ins_txt_id =
		gtk_signal_connect(GTK_OBJECT(document->textbox), "insert_text",
						   GTK_SIGNAL_FUNC(doc_insert_text_cb), document);
	document->del_txt_id =
		gtk_signal_connect(GTK_OBJECT(document->textbox), "delete_text",
						   GTK_SIGNAL_FUNC(doc_delete_text_cb), document);

	/* set this new document as active notebook page */
	flush_queue();
	DEBUG_MSG("doc_new, before notebook_set_page, after flush_queue\n");
	gtk_notebook_set_page(GTK_NOTEBOOK(main_v->notebook),
						  g_list_length(main_v->documentlist) - 1);
	notebook_changed();
	DEBUG_MSG("doc_new, ended\n");

	return document;
}

void doc_destroy(Tdocument * doc)
{
	DEBUG_MSG("doc_destroy, started, notebook=%p, document=%p\n",
			  main_v->notebook, doc);
	DEBUG_MSG("doc_destroy, textbox=%p\n", doc->textbox);
#ifdef DEBUG
	if (doc->filename) {
		g_print("doc_destroy, filename=%s, not modified, destroy it\n",
				doc->filename);
	} else {
		g_print("doc_destroy, filename=NULL, not modified, destroy it\n");
	}

	DEBUG_MSG("doc_destroy, page num should be %d\n",
			  gtk_notebook_page_num(GTK_NOTEBOOK(main_v->notebook),
									doc->textbox->parent));

#endif
	gtk_notebook_remove_page(GTK_NOTEBOOK(main_v->notebook),
							 gtk_notebook_page_num(GTK_NOTEBOOK
												   (main_v->notebook),
												   doc->textbox->parent));
	DEBUG_MSG("doc_destroy, g_list_length(documentlist)=%d\n",
			  g_list_length(main_v->documentlist));

	if (g_list_length(main_v->documentlist) > 1) {
		main_v->documentlist = g_list_remove(main_v->documentlist, doc);
	} else {
		main_v->documentlist = g_list_remove(main_v->documentlist, doc);
		DEBUG_MSG
			("doc_destroy, last document removed from documentlist\n");
		g_list_free(main_v->documentlist);
		DEBUG_MSG("doc_destroy, freed documentlist\n");
		main_v->documentlist = NULL;
		DEBUG_MSG("doc_destroy, documentlist = NULL\n");
	}
	DEBUG_MSG("doc_destroy, g_list_length(documentlist)=%d\n",
			  g_list_length(main_v->documentlist));
	if (doc->filename) {
		DEBUG_MSG("doc_destroy, doc->filename=%s\n", doc->filename);
		g_free(doc->filename);
	}
	g_free(doc);
/*		gtk_notebook_set_page(GTK_NOTEBOOK(main_v->notebook), ((gint) g_list_length(main_v->documentlist) - 1)); */
	notebook_changed();
}

void doc_save(Tdocument * doc, gint do_save_as, gint do_move)
{
	gchar *oldfilename = NULL;
#ifdef DEBUG
	g_assert(doc);
#endif

	DEBUG_MSG("doc_save, doc=%p, save_as=%d, do_move=%d\n", doc, do_save_as, do_move);
	if (doc->filename == NULL) {
		do_save_as = 1;
	}
	if (do_move) {
		do_save_as = 1;
	}

	if (do_save_as) {
		gchar *newfilename = NULL;
		gint index;
		statusbar_message(_("Save as..."), 1);
		oldfilename = doc->filename;
		doc->filename = NULL;
		newfilename = return_file(NULL);
		index = find_filename_in_documentlist(newfilename);
		DEBUG_MSG("doc_save, index=%d, filename=%p\n", index, newfilename);
#ifdef DEBUG
		if (newfilename) {
			g_print("doc_save, filename=%s\n", newfilename);
		}
#endif
		if (index != -1) {
			gchar *tmpstr;
			gint retval;
			gchar *options[] = {N_("Overwrite"), N_("Cancel"), NULL};
			tmpstr = g_strdup_printf(_("File %s is open, overwrite?"), newfilename);
			retval = multi_button_dialog(_("Bluefish warning, file is open"), 1, tmpstr, options);
			g_free(tmpstr);
			if (retval == 1) {
				g_free(newfilename);
				doc->filename = oldfilename;
				return;
			} else {
				Tdocument *tmpdoc;
				tmpdoc = (Tdocument *)g_list_nth_data(main_v->documentlist, index);
				DEBUG_MSG("doc_save, tmpdoc=%p\n", tmpdoc);
#ifdef DEBUG
				g_assert(tmpdoc);
				g_assert(tmpdoc->filename);
#endif
				g_free(tmpdoc->filename);
				tmpdoc->filename = NULL;
				tmpdoc->modified = 1;
				tmpstr = g_strconcat(_("Previously: "), strip_filename(newfilename), NULL);
				gtk_label_set(GTK_LABEL(tmpdoc->tab_label),tmpstr);
				g_free(tmpstr);
			}
		}
		doc->filename = newfilename;
	}

	if (doc->filename == NULL) {
		doc->filename = oldfilename;
		error_dialog(_("Bluefish error"), _("No filename known"));
		return;
	} 
	DEBUG_MSG("doc_save, returned file %s\n", doc->filename);
	if (do_save_as && oldfilename && main_v->props.link_management) {
		update_filenames_in_file(doc, oldfilename, doc->filename, 1);
	}
	statusbar_message(_("Save in progress"), 1);
	doc_textbox_to_file(doc, doc->filename);
	gtk_label_set(GTK_LABEL(doc->tab_label),
				  strip_filename(doc->filename));
	if (oldfilename) {
		if (do_move) {
			all_documents_update_links(doc, oldfilename,
									   doc->filename);
			unlink(oldfilename);
		}
		g_free(oldfilename);
	}
}

void doc_new_with_file(gchar * filename)
{

	Tdocument *doc;
	
	if ((filename == NULL) || (!file_exists_and_readable(filename))) {
		statusbar_message(_("Unable to open file"), 2);
		return;
	}
	if (!main_v->props.allow_multi_instances) {
		gint index;
		index = find_filename_in_documentlist(filename);
		if (index != -1){
			gtk_notebook_set_page(GTK_NOTEBOOK(main_v->notebook), index);
			return;
		}
	} 
	DEBUG_MSG("file_open, filename=%s exists\n", filename);
	doc = doc_new();
	doc->filename = g_strdup(filename);
	doc_file_to_textbox(doc, doc->filename);
	gtk_label_set(GTK_LABEL(doc->tab_label),
					  strip_filename(doc->filename));
}

void docs_new_from_files(GList * file_list)
{

	GList *tmplist;

	tmplist = g_list_first(file_list);
	while (tmplist) {
		doc_new_with_file((gchar *) tmplist->data);
		add_to_recent_list((gchar *) tmplist->data);
		tmplist = g_list_next(tmplist);
	}
}

/*
 * Function: close_save_cancel_dialog
 * Arguments:
 * 	title - title string
 * 	label - label string
 * Return value:
 * 	1 - close, 2 - save, 3 - cancel
 * Description:
 * 	create dialog which returns a close, save or cancel
 */
static gint close_save_cancel_dialog(gchar * title, gchar * label) {

	gint retval;
	gchar *buttons[] = {N_("save"), N_("close"), N_("cancel"), NULL};

	retval = multi_button_dialog(title, 2, label, buttons);
	
	switch (retval) {
		case 0:
			return 2;
		break;
		case 1:
			return 1;
		break;
		case 2:
			return 3;
		break;
	}
	return 3;
}


/*
returning 0 --> cancel or abort
returning 1 --> ok, closed or saved & closed
*/
gint doc_close(Tdocument * doc, gint warn_only)
{
	gchar *text;
	gint retval;

	if (!doc) {
		return 0;
	}

	if (!doc->modified && !doc->filename
		&& (gtk_text_get_length(GTK_TEXT(doc->textbox)) == 0)
		&& (g_list_length(main_v->documentlist) == 1)) {
		return 0;
	}

	if (doc->modified) {
		if (doc->filename) {
			text =
				g_strdup_printf(_("Are you sure you want to close\n%s ?"),
								doc->filename);
		} else {
			text =
				g_strdup(_
						 ("Are you sure you want to close\nthis untitled file ?"));
		}
		retval =
			close_save_cancel_dialog(_
									 ("Bluefish warning: file is modified!"),
text);
		g_free(text);
		switch (retval) {
		case 3:
			DEBUG_MSG("doc_close, retval=3, returning\n");
			return 0;
			break;
		case 2:
			doc_save(doc, 0, 0);
			if (doc->modified == 1) {
				/* something went wrong it's still not saved */
				return 0;
			}
			if (!warn_only) {
				doc_destroy(doc);
			}
			break;
		case 1:
			if (!warn_only) {
				doc_destroy(doc);
			}
			break;
		default:
			return 0;			/* something went wrong */
			break;
		}
	} else {
		DEBUG_MSG("doc_close, closing doc=%p\n", doc);
		if (!warn_only) {
			doc_destroy(doc);
		}
	}

	notebook_changed();
	return 1;
}

void file_save_cb(GtkWidget * widget, gpointer data)
{
	DEBUG_MSG("file_save_cb, file save started\n");
	doc_save(main_v->current_document, 0, 0);
}


void file_save_as_cb(GtkWidget * widget, gpointer data)
{
	DEBUG_MSG("file_save_as_cb, file save started\n");
	doc_save(main_v->current_document, 1, 0);
}

void file_move_to_cb(GtkWidget * widget, gpointer data)
{
	DEBUG_MSG("file_move_to_cb, file save started\n");
	doc_save(main_v->current_document, 1, 1);
}

void file_open_cb(GtkWidget * widget, gpointer data)
{

	GList *tmplist;
	DEBUG_MSG("file_open_cb, started\n");

	tmplist = return_files(NULL);
	if (!tmplist) {
		return;
	}
	docs_new_from_files(tmplist);
	free_stringlist(tmplist);
}

void file_insert_cb(GtkWidget * widget, gpointer data)
{
	gchar *tmpfilename;
	gint currentp;

	tmpfilename = return_file(NULL);
	if (tmpfilename == NULL) {
		error_dialog(_("Bluefish error"), _("No filename known"));
		return;
	} else {
		currentp =
			gtk_editable_get_position(GTK_EDITABLE
									  (main_v->current_document->textbox));
		gtk_text_set_point(GTK_TEXT(main_v->current_document->textbox),
						   currentp);
		doc_file_to_textbox(main_v->current_document, tmpfilename);
		g_free(tmpfilename);
	}
}

void file_new_cb(GtkWidget * widget, gpointer data)
{
	Tdocument *doc;
	/* This is the callback function for a file new button */
	DEBUG_MSG("file_new_cb, started\n");
	doc = doc_new();
	if (file_exists_and_readable(main_v->current_project.template) == 1) {
		doc_file_to_textbox(doc, main_v->current_project.template);
	}
	DEBUG_MSG
		("file_new_cb, end, current_document=%p, current_document->filename=%p\n",
		 main_v->current_document, main_v->current_document->filename);
}

void file_close_cb(GtkWidget * widget, gpointer data)
{
	DEBUG_MSG("file_close_cb, starting doc_close with %p\n",
			  main_v->current_document);
	doc_close(main_v->current_document, 0);
	DEBUG_MSG("file_close_cb, finished\n");
}

void file_close_all_cb(GtkWidget * widget, gpointer data)
{
	GList *tmplist;
	Tdocument *tmpdoc;
	gint retval = -1;

	DEBUG_MSG("file_close_all_cb, started\n");

	/* first a warning loop */
	if (test_docs_modified(NULL)) {
		gchar *options[] = {N_("save all"), N_("close all"), N_("choose per file"), N_("cancel"), NULL};
		retval =	multi_button_dialog(_("Bluefish warning: some file(s) are modified!"), 3, _("Some file(s) are modified\nplease choose your action"), options);
		if (retval == 3) {
			DEBUG_MSG("file_close_all_cb, cancel clicked, returning 0\n");
			return;
		}
	} else {
		retval = 1;
	}
	DEBUG_MSG("file_close_all_cb, after the warnings, retval=%d, now close all the windows\n", retval);

	tmplist = g_list_first(main_v->documentlist);
	while (tmplist) {
		tmpdoc = (Tdocument *) tmplist->data;
		if (test_only_empty_doc_left(tmpdoc)) {
			return;
		}
		
		switch (retval) {
		case 0:
			doc_save(tmpdoc, 0, 0);
			doc_destroy(tmpdoc);
			tmplist = g_list_first(main_v->documentlist);
		break;
		case 1:
			doc_destroy(tmpdoc);
			tmplist = g_list_first(main_v->documentlist);
		break;
		case 2:
			if (doc_close(tmpdoc, 0)) {
				tmplist = g_list_first(main_v->documentlist);
			} else {
				notebook_changed();
				return;
			}
		break;
		default:
			notebook_changed();
			return;
		break;
		}
	}
	notebook_changed();
	DEBUG_MSG("file_close_all_cb, finished\n");
}


/*
 * Function: file_save_all_cb
 * Arguments:
 * 	widget	- callback widget
 * 	data	- data for callback function
 * Return value:
 * 	void
 * Description:
 * 	Save all editor notebooks
 */
void file_save_all_cb(GtkWidget * widget, gpointer data)
{

	GList *tmplist;
	Tdocument *tmpdoc;

	tmplist = g_list_first(main_v->documentlist);
	while (tmplist) {
		tmpdoc = (Tdocument *) tmplist->data;
#ifdef DEBUG
		g_print("file_save_all_cb, a NULL document in the documentlist");
		g_assert(tmpdoc);
#endif
		if (tmpdoc->modified) {
			doc_save(tmpdoc, 0, 0);
		}
		tmplist = g_list_next(tmplist);
	}
}


/******************************************************************************/
/* this function destroys the current_document */
void destroy_current_document(void)
{
	if (main_v->current_document) {
		doc_destroy(main_v->current_document);
	} else {
		DEBUG_MSG
			("destroy_current_document, cannot close a NULL current_document\n");
	}
}


/*******************************************************************/
/*               Printing sysem                                    */
/*******************************************************************/

static void print_dialog_destroy_lcb(GtkWidget * widget, GdkEvent *event, gpointer data) {
	window_destroy(((Tprintwin *) data)->print_window);
	g_free(((Tprintwin *) data));
}

static void print_dialog_cancel_lcb(GtkWidget * widget, gpointer data) {
	print_dialog_destroy_lcb(NULL, NULL, data);
}

static char *temporary_print_file(Tdocument * doc, Tprintwin * pw)
{
	char *filename;

	filename =
		g_strdup_printf("%s/.bluefish_print.%d%d",
						gtk_entry_get_text(GTK_ENTRY(pw->temp_dir)),
						(int) time(NULL), getpid());
	doc_textbox_to_file(main_v->current_document, filename);
	return filename;
}


static void execute_print(GtkWidget * widget, gpointer data)
{
	char *orig_entry, *fname, *final_command, error_msg[256];
	DIR *dp;					/* to check if the temporary dir exists */
	Tprintwin *pw = (Tprintwin *) data;

	/* If no entry was given, there is no reason to continue */
	orig_entry =
		gtk_editable_get_chars(GTK_EDITABLE(pw->print_cmd), 0, -1);
	if (strlen(orig_entry) <= 0) {
		error_dialog(_("Error"), _("No commandline set!"));
		g_free(orig_entry);
		print_dialog_destroy_lcb(NULL, NULL, data);
		return;
	}

	/* Check if temporary directory exists */
	if ((dp = opendir(gtk_entry_get_text(GTK_ENTRY(pw->temp_dir)))) ==
		NULL) {
		strcpy(error_msg, gtk_entry_get_text(GTK_ENTRY(pw->temp_dir)));
		strcat(error_msg, _(" doesn't exist or isn't a directory."));
		error_dialog(_("Error"), error_msg);
		g_free(orig_entry);
		print_dialog_destroy_lcb(NULL, NULL, data);
		return;
	}
	closedir(dp);

	/* Now, let's explain a bit how this function works.
	 * We must create a temporary file in which the contents of
	 * the text area will be saved, in order to be printed. During
	 * this procedure, the actual document isn't saved. Then,
	 * we must replace %s in the command line with the temporary
	 * file name. We execute the command line and then erase
	 * the temporary file */

	/* Create temporary file and save its name at fname */
	if (!(fname = temporary_print_file(main_v->current_document, pw))) {
		/* if it couldn't be created, cancel print */
		error_dialog(_("Error"),
					 _("Error in creating temporary print file."));
		g_free(orig_entry);
		print_dialog_destroy_lcb(NULL, NULL, data);
		return;
	}

	/* Build the command and execute it. */
	final_command = g_strdup_printf(orig_entry, fname);

	if (system(final_command) == -1) {
		error_dialog(_("Error"),
					 _
					 ("System error in printing file. File not printed."));
	}

	/* Erase the temporary file */
	if (unlink(fname)) {
		error_dialog(_("Error"), _("Unable to erase temporary file."));
	}

	g_free(fname);
	g_free(final_command);
	g_free(orig_entry);
	print_dialog_destroy_lcb(NULL, NULL, data);
}

static void show_print_dialog()
{
	GtkWidget *button;
	GtkWidget *vbox, *hbox;
	Tprintwin *pw;

	pw = g_malloc(sizeof(Tprintwin));
	/* Create window */
	pw->print_window =
		window_full(_("Print Dialog"), GTK_WIN_POS_MOUSE,
					GTK_WINDOW_DIALOG, 3,print_dialog_destroy_lcb , pw);

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(pw->print_window), vbox);

	gtk_box_pack_start(GTK_BOX(vbox),
					   gtk_label_new(_
									 ("Enter print command-line below.\n Remember that you must place %s where\n the filename is supposed to be placed. \n\nPrint command:")),
					   TRUE, TRUE, 0);
	pw->print_cmd = entry_with_text("lpr %s", 255);
	gtk_box_pack_start(GTK_BOX(vbox), pw->print_cmd, TRUE, TRUE, 0);

	gtk_box_pack_start(GTK_BOX(vbox),
					   gtk_label_new(_("Temporary directory:")), TRUE,
					   TRUE, 0);
	pw->temp_dir = entry_with_text("/tmp", 255);
	gtk_box_pack_start(GTK_BOX(vbox), pw->temp_dir, TRUE, TRUE, 0);

	/* Ok and Cancel buttons */
	hbox = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);

	button = bf_stock_button(_(" Print "), execute_print, pw);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);

	button = bf_stock_cancel_button(print_dialog_cancel_lcb, pw);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);

	gtk_widget_show_all(pw->print_window);
}

void file_print_cb(GtkWidget * widget, gpointer data)
{
	show_print_dialog();
}
